from uuid import UUID
from typing import Dict
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm.util import AliasedClass
from .alchemy_orm import make_aliased_mapper


class AliasedManager:
    __aliased_map: Dict[UUID, Dict[str, AliasedClass]] = dict()

    @classmethod
    def get_or_create(
        cls,
        uuid: UUID,
        mapper: DeclarativeMeta,
        field_name: str
    ) -> AliasedClass:
        if cls.is_exist(uuid=uuid, field_name=field_name):
            return cls.__aliased_map[uuid][field_name]

        aliased_mapper = make_aliased_mapper(
            mapper=mapper,
            field_name=field_name
        )
        cls.__aliased_map[uuid] = cls.__aliased_map.get(uuid, {})
        cls.__aliased_map[uuid][field_name] = aliased_mapper

        return aliased_mapper

    @classmethod
    def delete(
        cls,
        uuid: UUID,
    ):
        cls.__aliased_map.pop(uuid, None)

    @classmethod
    def is_exist(
        cls,
        uuid: UUID,
        field_name: str,
    ) -> bool:
        return bool(cls.__aliased_map.get(uuid, {}).get(field_name, False))



from typing import Any, Callable, Optional, Union
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from sqlalchemy.sql.expression import BinaryExpression
from .base import BaseClauseBinder


class SelfMethodBinder(BaseClauseBinder):
    def _is_self_method(
        self,
        lookup: str,
    ) -> bool:
        return self._get_self_method(lookup=lookup) is not None

    def _get_self_method(
        self,
        lookup: str
    ) -> Optional[Callable[
        [str, Any, DeclarativeMeta, Select], BinaryExpression]
    ]:
        lookup_parts = lookup.split(self.LOOKUP_STRING)

        for i in range(len(lookup_parts)-1, -1, -1):
            self_method = getattr(
                self,
                self.LOOKUP_STRING.join(lookup_parts[:i]),
                None
            )
            if self_method is not None:
                return self_method

    def _bind_self_method(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
    ) -> Select:
        self_method = self._get_self_method(lookup=lookup)
        if self_method is None:
            return stmt

        expression = self_method(lookup, value, mapper, stmt)

        return stmt.where(expression)



from abc import abstractmethod
from sqlalchemy import func, distinct
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, select
from ..clause_binder import ClauseBinder
from ..query import CRUDQuery
from .select_provider import SelectProvider


class CountProvider(SelectProvider):
    @abstractmethod
    async def select_count(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_count_stmt(self, *args, **kwargs):
        pass

    def _make_count_stmt(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder
    ) -> Select:
        """
        select count(distinct alias1.id)
        from (
            select test.id as id, test.name as name,
            test2.id as test2_id, test2.test_id as test2_test_id
            from test join test2 on test.id = test2.test_id
            order by test.name, test2.id
        ) as alias1
        """
        select_stmt = self._make_select_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder
        )
        select_count_stmt = select(func.count()).select_from(select_stmt)

        subquery = select_stmt.subquery()
        for column in subquery.columns:
            if not (column.primary_key or column.unique):
                continue
            if not hasattr(mapper, column.name):
                continue

            select_count_stmt = select(func.count(distinct(column)))
            break

        return select_count_stmt

    async def _select_count(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
    ) -> int:
        select_count_stmt = self._make_count_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
        )

        return await self._session.scalar(select_count_stmt)



from __future__ import annotations
from typing import Type, List
from sqlalchemy.engine.row import Row
from sqlalchemy.orm import DeclarativeMeta
from .base import BaseQuery
from .select_query import SelectQuery
from .insert_query import InsertQuery
from .update_query import UpdateQuery
from .delete_query import DeleteQuery


class CRUDQuery(
    InsertQuery,
    SelectQuery,
    UpdateQuery,
    DeleteQuery,
    BaseQuery
):
    pass

    @classmethod
    def get_field_query(cls, field_name: str) -> Type[CRUDQuery]:
        return super().get_field_query(field_name=field_name)

    @classmethod
    def from_returning_mapper(
        cls,
        mapper: DeclarativeMeta
    ) -> CRUDQuery:
        return super().from_returning_mapper(mapper=mapper)

    @classmethod
    def from_returning_mappers(
        cls,
        mappers: List[DeclarativeMeta]
    ) -> List[CRUDQuery]:
        return super().from_returning_mappers(mappers=mappers)

    @classmethod
    def from_selected_row(
        cls,
        row: Row
    ) -> CRUDQuery:
        return super().from_selected_row(row=row)

    @classmethod
    def from_selected_rows(
        cls,
        rows: List[Row]
    ) -> List[CRUDQuery]:
        return super().from_selected_rows(rows=rows)



"""
Implements cls or instance decorator
https://stackoverflow.com/questions/28237955/same-name-for-classmethod-and-instancemethod#:~:text=Class%20and%20instance%20methods%20live,will%20win%20in%20that%20case.&text=This%20is%20explicitly%20documented%3A,instance%20(such%20as%20C().
"""


class cls_or_ins(classmethod):
    def __get__(self, instance, type_):
        descr_get = super().__get__ \
            if instance is None else self.__func__.__get__
        return descr_get(instance, type_)



from typing import Union, Type
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql import Select, Insert, Update, Delete
from ..query import CRUDQuery
from .base import BaseProvider


class JoinProvider(BaseProvider):
    def _join(
        self,
        field_name: str,
        query: Union[CRUDQuery, Type[CRUDQuery]],
        stmt: Union[Select, Insert, Update, Delete],
        mapper: DeclarativeMeta,
        aliased_mapper: AliasedClass
    ) -> Union[Select, Insert, Update, Delete]:
        mapper_field = getattr(mapper, field_name)
        join_strategy = query.get_join_strategy(field_name=field_name)

        stmt2 = stmt.join(
            aliased_mapper,
            onclause=mapper_field,
            isouter=join_strategy.is_outer,
            full=join_strategy.is_full
        )

        return stmt2



from abc import abstractmethod
from uuid import UUID
from typing import Dict, Any, Optional, Union, Sequence
from sqlalchemy import and_, or_
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql import Select, Insert, Update, Delete
from sqlalchemy.sql.expression import BinaryExpression
from ..utils import is_relationship, get_column, AliasedManager


class BaseClauseBinder:
    LOOKUP_STRING = '__'
    OR_OPERATOR = '_or_'
    AND_OPERATOR = '_and_'

    @abstractmethod
    def bind(self, *args, **kwargs):
        pass

    @abstractmethod
    def _is_self_method(
        self,
        lookup: str
    ):
        pass

    @abstractmethod
    def _bind_self_method(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
    ) -> Select:
        pass

    @abstractmethod
    def _bind_string_expression_method(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
        uuid: UUID,
    ) -> Select:
        pass

    @abstractmethod
    def _get_expression(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        uuid: UUID,
    ):
        pass

    def _bind(
        self,
        clause: Dict[str, Any],
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
        uuid: UUID,
    ) -> Select:
        for lookup, value in clause.items():
            stmt = self._bind_expressions(
                lookup=lookup,
                value=value,
                mapper=mapper,
                stmt=stmt,
                uuid=uuid,
            )

        return stmt

    def _bind_expressions(
        self,
        lookup: str,
        value: Any,
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
        uuid: UUID,
    ) -> Select:
        if self._is_self_method(lookup=lookup):
            return self._bind_self_method(
                lookup=lookup,
                value=value,
                mapper=mapper,
                stmt=stmt
            )

        if type(value) is dict:
            stmt = self._bind_dict_value(
                lookup=lookup,
                value=value,
                mapper=mapper,
                stmt=stmt,
                uuid=uuid,
            )

        return self._bind_string_expression_method(
            lookup=lookup,
            value=value,
            mapper=mapper,
            stmt=stmt,
            uuid=uuid,
        )

    def _bind_dict_value(
        self,
        lookup: str,
        value: Dict[str, Any],
        mapper: DeclarativeMeta,
        stmt: Union[Select, Insert, Update, Delete],
        uuid: UUID,
    ) -> Select:
        expression = self._get_dict_expression(
            lookup=lookup,
            value=value,
            mapper=mapper,
            uuid=uuid,
        )
        if expression is None:
            return stmt

        if isinstance(expression, Sequence):
            return stmt.where(*expression)

        return stmt.where(expression)

    def _get_dict_expression(
        self,
        lookup: str,
        value: Dict[str, Any],
        mapper: DeclarativeMeta,
        uuid: UUID,
    ) -> Optional[Union[BinaryExpression, Sequence[BinaryExpression]]]:
        if lookup in (self.AND_OPERATOR, self.OR_OPERATOR) \
                and type(value) is not dict:
            raise ValueError(
                f'If lookup in {(self.AND_OPERATOR, self.OR_OPERATOR)}, so'
                f'value have to be dict'
            )

        if lookup == self.OR_OPERATOR:
            return self._get_or_expression(
                clause=value,
                mapper=mapper,
                uuid=uuid,
            )

        if lookup == self.AND_OPERATOR:
            return self._get_and_expression(
                clause=value,
                mapper=mapper,
                uuid=uuid,
            )

        column = get_column(mapper=mapper, field_name=lookup)
        if column is not None:
            if is_relationship(mapper_field=column):
                return self._get_expressions(
                    clause=value,
                    mapper=AliasedManager.get_or_create(
                        uuid=uuid,
                        mapper=mapper,
                        field_name=lookup
                    ),
                    uuid=uuid,
                )

    def _get_or_expression(
        self,
        clause: Dict[str, Any],
        mapper: DeclarativeMeta,
        uuid: UUID,
    ) -> BinaryExpression:
        expressions = self._get_expressions(
            clause=clause,
            mapper=mapper,
            uuid=uuid,
        )
        return or_(*expressions)

    def _get_and_expression(
        self,
        clause: Dict[str, Any],
        mapper: DeclarativeMeta,
        uuid: UUID,
    ) -> BinaryExpression:
        expressions = self._get_expressions(
            clause=clause,
            mapper=mapper,
            uuid=uuid,
        )
        return and_(*expressions)

    def _get_expressions(
        self,
        clause: Dict[str, Any],
        mapper: Union[DeclarativeMeta, AliasedClass],
        uuid: UUID,
    ) -> Sequence[BinaryExpression]:
        expressions = []
        for lookup, value in clause.items():
            if type(value) is dict:
                expressions.append(self._get_dict_expression(
                    lookup=lookup,
                    value=clause,
                    mapper=mapper,
                    uuid=uuid,
                ))
                continue

            expressions.append(self._get_expression(
                lookup=lookup,
                value=value,
                mapper=mapper,
                uuid=uuid,
            ))

        return expressions



from __future__ import annotations
from abc import abstractmethod
from uuid import uuid4, UUID
from typing import Union, Type, List, Optional
from sqlalchemy import select
from sqlalchemy.sql import Select, Insert, Update
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty, \
    RelationshipProperty, InstrumentedAttribute, aliased
from sqlalchemy.orm.util import AliasedClass
from ..clause_binder import ClauseBinder
from ..query import CRUDQuery, FIELD_NAME_SEPARATOR
from ..utils import AliasedManager
from .base import BaseProvider
from .join_provider import JoinProvider
from .pagination_provider import PaginationProvider
from .sorting_provider import SortingProvider


class SelectProvider(
    JoinProvider,
    PaginationProvider,
    SortingProvider,
    BaseProvider
):
    @abstractmethod
    async def select(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_select_stmt(self, *args, **kwargs):
        pass

    def _make_select_stmt(
        self,
        query: Union[Type[CRUDQuery], CRUDQuery],
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        uuid: UUID = None,
    ) -> Select:
        uuid = uuid or uuid4()
        select_stmt = self._make_simple_select_stmt(
            query=query,
            mapper=mapper,
            uuid=uuid,
        )

        select_stmt = self.bind_pagination(
            query=query,
            select_stmt=select_stmt
        )

        select_stmt = self.bind_sorting(
            query=query,
            mapper=mapper,
            select_stmt=select_stmt,
        )

        if query.is_instance():
            select_stmt = self._bind_clause(
                clause=query.get_filters(),
                mapper=mapper,
                stmt=select_stmt,
                clause_binder=clause_binder,
                uuid=uuid,
            )

        AliasedManager.delete(uuid=uuid)

        return select_stmt

    def _make_select_from_stmt(
        self,
        query: CRUDQuery,
        stmt: Union[Insert, Update],
        mapper: DeclarativeMeta,
        uuid: Optional[UUID] = None,
    ) -> Select:
        uuid = uuid or uuid4()

        selectable_columns = self._get_selectable_columns(
            mapper=mapper,
        )
        returning_cte_alias = stmt.returning(*selectable_columns).cte().alias()
        aliased_mapper = aliased(mapper, alias=returning_cte_alias)

        select_stmt = self._make_simple_select_stmt(
            query=query.get_class(),
            mapper=aliased_mapper,
            uuid=uuid
        )

        AliasedManager.delete(uuid=uuid)

        return select_stmt

    def _make_select_from_insert(
        self,
        query: CRUDQuery,
        insert_stmt: Insert,
        mapper: DeclarativeMeta,
    ) -> Select:
        """
        with inserted_cte as (
            insert into test(id, name) values
            (1, 'some_name')
            returning id
        )
        select test.*, test2.*
        from test
        join inserted_cte on test.id = inserted_cte.id
        join test2 on test.test2_id = test2.id
        """
        return self._make_select_from_stmt(
            query=query,
            stmt=insert_stmt,
            mapper=mapper,
        )

    def _make_select_from_update(
        self,
        query: CRUDQuery,
        update_stmt: Update,
        mapper: DeclarativeMeta,
        uuid: UUID,
    ) -> Select:
        """
        with updated_cte as (
            update test
            set name = 'some_new_name'
            where id >= 1
            returning id
        )
        select test.*, test2.*
        from test
        join updated_cte on test.id = updated_cte.id
        join test2 on test.test2_id = test2.id
        """
        return self._make_select_from_stmt(
            query=query,
            stmt=update_stmt,
            mapper=mapper,
            uuid=uuid,
        )

    def _make_simple_select_stmt(
        self,
        query: Type[CRUDQuery],
        mapper: Union[DeclarativeMeta, AliasedClass],
        uuid: UUID,
        label_prefix: Optional[str] = None,
        select_stmt: Optional[Select] = None,
    ) -> Select:

        if select_stmt is None:
            select_stmt = select()

        type_hints = query.get_type_hints()

        for field_name, type_hint in type_hints.items():
            mapper_field = getattr(mapper, field_name, None)
            if mapper_field is None:
                continue

            if isinstance(mapper_field.property, ColumnProperty):
                select_stmt = select_stmt.add_columns(
                    mapper_field.label(
                        self._make_column_label(
                            mapper_field=mapper_field,
                            label_prefix=label_prefix
                        )
                    )
                )
                continue

            if isinstance(mapper_field.property, RelationshipProperty):
                aliased_mapper = AliasedManager.get_or_create(
                    uuid=uuid,
                    mapper=mapper,
                    field_name=field_name
                )

                select_stmt = self._join(
                    field_name=field_name,
                    query=query,
                    stmt=select_stmt,
                    mapper=mapper,
                    aliased_mapper=aliased_mapper,
                )

                select_stmt = self._make_simple_select_stmt(
                    select_stmt=select_stmt,
                    query=query.get_field_query(field_name),
                    uuid=uuid,
                    mapper=aliased_mapper,
                    label_prefix=field_name,
                )

        return select_stmt

    @staticmethod
    def _get_selectable_columns(
        mapper: DeclarativeMeta
    ) -> List[InstrumentedAttribute]:
        columns = list()
        for column_name, column in mapper.__dict__.items():
            if not isinstance(column, InstrumentedAttribute):
                continue

            if not isinstance(column.property, ColumnProperty):
                continue

            columns.append(column)

        return columns

    @staticmethod
    def _make_column_label(
        mapper_field: InstrumentedAttribute,
        label_prefix: Optional[str] = None,
    ) -> str:
        if label_prefix is None:
            return mapper_field.name

        return label_prefix + FIELD_NAME_SEPARATOR + mapper_field.name

    async def _select(
        self,
        query: Union[Type[CRUDQuery], CRUDQuery],
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder
    ) -> List[CRUDQuery]:
        select_stmt = self._make_select_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder
        )

        scalar_result = await self._session.execute(select_stmt)

        return query.from_selected_rows(rows=scalar_result.all())



from sqlalchemy.orm import DeclarativeMeta, ColumnProperty
from sqlalchemy.sql import Select, nullsfirst, nullslast
from ..query import CRUDQuery
from .base import BaseProvider


class SortingProvider(BaseProvider):
    def bind_sorting(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        select_stmt: Select
    ) -> Select:
        if query.order_by is None:
            return select_stmt

        mapper_field = getattr(mapper, query.order_by, None)
        if mapper_field is None:
            return select_stmt

        if mapper_field.property is not ColumnProperty:
            return select_stmt

        sorting_column = mapper_field.desc() \
            if query.reversed else mapper_field.ack()

        if query.nulls_place is None:
            return select_stmt.order_by(sorting_column)

        if query.nulls_place == 'first':
            return select_stmt.order_by(nullsfirst(sorting_column))

        if query.nulls_place == 'last':
            return select_stmt.order_by(nullslast(sorting_column))

        return select_stmt



from abc import ABC
from typing import List, Dict, Any, Optional, Sequence, Type
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from ..clause_binder import ClauseBinder
from ..query import CRUDQuery
from .select_provider import SelectProvider
from .insert_provider import InsertProvider
from .update_provider import UpdateProvider
from .delete_provider import DeleteProvider
from .count_provider import CountProvider


class AbstractProvider(
    ABC,
    CountProvider,
    InsertProvider,
    UpdateProvider,
    SelectProvider,
    DeleteProvider,
):
    _mapper: DeclarativeMeta
    _clause_binder: ClauseBinder
    _query_type: Type[CRUDQuery]

    def make_select_stmt(
        self,
        **kwargs
    ) -> Select:
        return self._make_select_stmt(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder
        )

    def make_count_stmt(
        self,
        **kwargs
    ) -> Select:
        return self._make_count_stmt(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder
        )

    def make_insert_stmt(
        self,
        **kwargs,
    ) -> Insert:
        return self._make_insert_stmt(
            query=self._query_type(**kwargs),
            mapper=self._mapper
        )

    def make_bulk_insert_stmt(
        self,
        values_seq: Sequence[Dict[str, Any]],
        returning: bool = True,
    ) -> Insert:
        return self._make_bulk_insert_stmt(
            query=self._query_type,
            values_seq=values_seq,
            mapper=self._mapper,
            returning=returning
        )

    def make_update_stmt(
        self,
        filters: Dict[str, Any],
        values: Dict[str, Any],
        returning: bool = True
    ) -> Update:
        return self._make_update_stmt(
            query=self._query_type.set_filters(**filters).set_values(**values),
            mapper=self._mapper,
            clause_binder=self._clause_binder,
            returning=returning
        )

    def make_update_stmt_from_kwargs(
        self,
        returning: bool = True,
        **kwargs
    ) -> Update:
        return self._make_update_stmt_from_kwargs(
            query=self._query_type,
            mapper=self._mapper,
            clause_binder=self._clause_binder,
            returning=returning,
            **kwargs
        )

    def make_delete_stmt(self, **kwargs) -> Delete:
        return self._make_delete_stmt(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder
        )

    async def select_count(
        self,
        **kwargs
    ) -> int:
        return await self._select_count(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder
        )

    async def select(
        self,
        **kwargs
    ) -> List[CRUDQuery]:
        return await self._select(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder
        )

    async def insert(
        self,
        returning: bool = True,
        **kwargs
    ) -> Optional[CRUDQuery]:
        return await self._insert(
            query=self._query_type(**kwargs),
            mapper=self._mapper,
            returning=returning
        )

    async def bulk_insert(
        self,
        values_seq: Sequence[Dict[str, Any]],
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        return await self._bulk_insert(
            query=self._query_type,
            values_seq=values_seq,
            mapper=self._mapper,
            returning=returning
        )

    async def update(
        self,
        returning: bool = True,
        **kwargs,
    ) -> Optional[Sequence[CRUDQuery]]:
        return await self._update(
            query=self._query_type,
            mapper=self._mapper,
            clause_binder=self._clause_binder,
            returning=returning,
            **kwargs
        )

    async def update_from_query(
        self,
        filters: Dict[str, Any],
        values: Dict[str, Any],
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        return await self._update_from_query(
            query=self._query_type().set_filters(**filters).set_values(**values),
            mapper=self._mapper,
            clause_binder=self._clause_binder,
            returning=returning
        )

    async def delete(
        self,
        **kwargs
    ):
        await self._delete(
            query=self._query_type.set_filters(**kwargs),
            mapper=self._mapper,
            clause_binder=self._clause_binder,
        )



from abc import abstractmethod
from uuid import uuid4
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import delete, Delete
from ..clause_binder import ClauseBinder
from ..query import CRUDQuery
from ..utils import AliasedManager
from .base import BaseProvider


class DeleteProvider(BaseProvider):
    @abstractmethod
    async def delete(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_delete_stmt(self, *args, **kwargs):
        pass

    def _make_delete_stmt(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder
    ) -> Delete:
        delete_stmt = delete(mapper)

        uuid = uuid4()
        delete_stmt = self._bind_clause(
            clause=query.get_filters(),
            mapper=mapper,
            stmt=delete_stmt,
            clause_binder=clause_binder,
            uuid=uuid,
        )
        AliasedManager.delete(uuid=uuid)

        return delete_stmt

    async def _delete(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder
    ):
        delete_stmt = self._make_delete_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
        )
        await self._session.execute(delete_stmt)

    def __make_delete_join_stmt(self):
        """
        delete
        from test2
        using test
        where test2.test_id = test.id and
        test.id = 1 and test2.id < 2;
        Will be in next realize
        """
        raise NotImplementedError



from typing import Optional, Literal
from .base import BaseQuery


class SortingQuery(BaseQuery):
    order_by: Optional[str] = None
    reversed: bool = False
    nulls_place: Literal['first', 'last', None] = None



from __future__ import annotations
from abc import abstractmethod
from uuid import UUID
from typing import Any, Dict, Union
from sqlalchemy.ext.asyncio import AsyncSession, async_scoped_session
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select, Insert, Update, Delete
from ..clause_binder import ClauseBinder


class BaseProvider:
    @property
    @abstractmethod
    def _session(self) -> Union[AsyncSession, async_scoped_session]:
        pass

    def _bind_clause(
        self,
        clause: Dict[str, Any],
        stmt: Union[Select, Insert, Update, Delete],
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        uuid: UUID,
    ) -> Union[Select, Insert, Update, Delete]:
        return clause_binder.bind(
            clause=clause,
            stmt=stmt,
            mapper=mapper,
            uuid=uuid,
        )



"""
Utils for work with SQLAlchemy orm
"""
from typing import Optional
from sqlalchemy.orm import DeclarativeMeta, InstrumentedAttribute, \
    ColumnProperty, RelationshipProperty, aliased
from sqlalchemy.orm.util import AliasedClass


def get_related_mapper(
    mapper: DeclarativeMeta,
    field_name: str,
) -> Optional[DeclarativeMeta]:
    """
    returns declarative orm model from relationship field
    """
    mapper_field: InstrumentedAttribute = getattr(mapper, field_name, None)
    if mapper_field is None:
        return

    return mapper_field.property.mapper.class_


def get_column(
    mapper: DeclarativeMeta,
    field_name: str,
) -> Optional[InstrumentedAttribute]:
    return getattr(mapper, field_name, None)


def is_column(
    mapper_field: InstrumentedAttribute
) -> bool:
    return isinstance(mapper_field.property, ColumnProperty)


def is_relationship(
    mapper_field: InstrumentedAttribute
) -> bool:
    return isinstance(mapper_field.property, RelationshipProperty)


def make_aliased_mapper(
    mapper: DeclarativeMeta,
    field_name: str,
) -> Optional[AliasedClass]:
    related_mapper = get_related_mapper(
        mapper=mapper,
        field_name=field_name
    )
    aliased_mapper = aliased(related_mapper, name=field_name)
    return aliased_mapper



from typing import List, Optional, Sequence
from sqlalchemy.orm import DeclarativeMeta
from ..clause_binder.clause_binder import ClauseBinder
from ..query import CRUDQuery
from .select_provider import SelectProvider
from .insert_provider import InsertProvider
from .update_provider import UpdateProvider
from .delete_provider import DeleteProvider
from .count_provider import CountProvider


class Provider(
    CountProvider,
    InsertProvider,
    UpdateProvider,
    SelectProvider,
    DeleteProvider,
):
    async def select(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
    ) -> List[CRUDQuery]:
        return await self._select(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
        )

    async def insert(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[CRUDQuery]:
        return await self._insert(
            query=query,
            mapper=mapper,
            returning=returning
        )

    async def bulk_insert(
        self,
        queries: Sequence[CRUDQuery],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        return await self._bulk_insert(
            queries=queries,
            mapper=mapper,
            returning=returning
        )

    async def update(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        return await self._update(
            query=query,
            mapper=mapper,
            returning=returning
        )

    async def delete(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta
    ):
        await self._delete(
            query=query,
            mapper=mapper
        )



from abc import abstractmethod
from uuid import uuid4
from typing import Any, Dict, Sequence, Optional, Union, Type
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty
from sqlalchemy.sql import update, Update
from ..clause_binder import ClauseBinder
from ..query import CRUDQuery
from ..utils import AliasedManager
from .base import BaseProvider
from .select_provider import SelectProvider


class UpdateProvider(SelectProvider, BaseProvider):
    @abstractmethod
    async def update(self, *args, **kwargs):
        pass

    @abstractmethod
    async def update_from_query(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_update_stmt(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_update_stmt_from_kwargs(self, *args, **kwargs):
        pass

    def _make_update_stmt(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        returning: bool = True
    ) -> Update:
        uuid = uuid4()

        update_stmt = update(mapper)
        update_stmt = self._bind_clause(
            clause=query.get_filters(),
            mapper=mapper,
            stmt=update_stmt,
            clause_binder=clause_binder,
            uuid=uuid,
        )

        updatable_values = self.__make_updatable_values(
            query=query,
            mapper=mapper,
        )
        if not updatable_values:
            raise ValueError(
                'Attr values is empty'
            )

        update_stmt = update_stmt.values(**updatable_values)

        result_stmt = update_stmt
        if returning:
            result_stmt = self._make_select_from_update(
                query=query,
                update_stmt=update_stmt,
                mapper=mapper,
                uuid=uuid,
            )

        AliasedManager.delete(uuid=uuid)

        return result_stmt

    @staticmethod
    def _query_from_kwargs(
        query: Union[Type[CRUDQuery], CRUDQuery],
        clause_binder: ClauseBinder,
        **kwargs
    ) -> CRUDQuery:
        result_query = query
        for field, value in kwargs.items():
            if clause_binder.LOOKUP_STRING in field:
                result_query = result_query.set_filters(**{field: value})
            else:
                result_query = result_query.set_values(**{field: value})

        return result_query

    def _make_update_stmt_from_kwargs(
        self,
        query: Union[Type[CRUDQuery], CRUDQuery],
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        returning: bool = True,
        **kwargs
    ) -> Update:
        query = self._query_from_kwargs(
            query=query,
            clause_binder=clause_binder,
            **kwargs
        )
        return self._make_update_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
            returning=returning
        )

    def __make_bulk_update_stmt(self):
        """
        Will be released in next version
        Use simple make_update_stmt for simple bulk_updates
        """
        raise NotImplementedError

    async def _update(
        self,
        query: Union[Type[CRUDQuery], CRUDQuery],
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        returning: bool = True,
        **kwargs,
    ) -> Optional[Sequence[CRUDQuery]]:
        """
        Returns sequence of query instance if returning is True
        """
        update_stmt = self._make_update_stmt_from_kwargs(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
            returning=returning,
            **kwargs
        )

        scalar_result = await self._session.execute(update_stmt)

        if not returning:
            return

        return query.from_selected_rows(scalar_result.all())

    async def _update_from_query(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        clause_binder: ClauseBinder,
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        update_stmt = self._make_update_stmt(
            query=query,
            mapper=mapper,
            clause_binder=clause_binder,
            returning=returning,
        )

        scalar_result = await self._session.execute(update_stmt)

        if not returning:
            return

        return query.from_selected_rows(scalar_result.all())

    @staticmethod
    def __make_updatable_values(
        query: CRUDQuery,
        mapper: DeclarativeMeta
    ) -> Dict[str, Any]:
        updatable_values = dict()

        for field_name, value in query.get_values().items():
            mapper_field = getattr(mapper, field_name, None)
            if mapper_field is None:
                continue

            if not isinstance(mapper_field.property, ColumnProperty):
                continue

            updatable_values[field_name] = value

        return updatable_values



from abc import abstractmethod
from typing import Any, Optional, List, Mapping, Sequence, Type, Union, Dict
from sqlalchemy.orm import DeclarativeMeta, ColumnProperty
from sqlalchemy.sql import insert, Insert, Select
from ..query import CRUDQuery
from .base import BaseProvider
from .select_provider import SelectProvider


class InsertProvider(SelectProvider, BaseProvider):
    @abstractmethod
    async def insert(self, *args, **kwargs):
        pass

    @abstractmethod
    async def bulk_insert(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_insert_stmt(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_bulk_insert_stmt(self, *args, **kwargs):
        pass

    def _make_insert_stmt(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Union[Insert, Select]:
        insert_stmt = insert(mapper)

        insertable_values = self.__make_insertable_values(
            mapper=mapper,
            values=query.dict,
        )

        insert_stmt = insert_stmt.values(**insertable_values)

        if returning:
            return self._make_select_from_insert(
                query=query,
                insert_stmt=insert_stmt,
                mapper=mapper,
            )

        return insert_stmt

    def _make_bulk_insert_stmt(
        self,
        query: Union[Type[CRUDQuery], CRUDQuery],
        values_seq: Sequence[Dict[str, Any]],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Union[Insert, Select]:
        """
        """
        insertable_values_seq: List[Mapping[str, Any]] = []

        for values in values_seq:
            insertable_values = self.__make_insertable_values(
                mapper=mapper,
                values=values
            )
            insertable_values_seq.append(insertable_values)

        insert_stmt = insert(mapper)
        insert_stmt = insert_stmt.values(insertable_values_seq)

        if returning:
            return self._make_select_from_insert(
                query=query,
                insert_stmt=insert_stmt,
                mapper=mapper,
            )

        return insert_stmt

    async def _insert(
        self,
        query: CRUDQuery,
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[CRUDQuery]:
        """
        if returning is True, returns instance of passed query
        """
        insert_stmt = self._make_insert_stmt(
            query=query,
            mapper=mapper,
            returning=returning
        )

        scalar_result = await self._session.execute(insert_stmt)

        if returning:
            return query.from_selected_row(scalar_result.first())

    async def _bulk_insert(
        self,
        query: Union[CRUDQuery, Type[CRUDQuery]],
        values_seq: Sequence[Dict[str, Any]],
        mapper: DeclarativeMeta,
        returning: bool = True,
    ) -> Optional[Sequence[CRUDQuery]]:
        """
        If returning is True, returns iterable object that contains
        CRUDQuery
        """
        if not values_seq:
            return

        insert_stmt = self._make_bulk_insert_stmt(
            query=query,
            values_seq=values_seq,
            mapper=mapper,
            returning=returning
        )

        scalar_result = await self._session.execute(insert_stmt)

        if returning:
            return query.from_selected_rows(scalar_result.all())

    @staticmethod
    def __make_insertable_values(
        mapper: DeclarativeMeta,
        values: Dict[str, Any],
    ) -> Mapping[str, Any]:
        insertable_values = dict()

        for field_name, value in values.items():
            mapper_field = getattr(mapper, field_name, None)

            if mapper_field is None:
                continue

            if not isinstance(mapper_field.property, ColumnProperty):
                continue

            insertable_values[field_name] = value

        return insertable_values
